package com.zanxiang.common.redis.aspect;

import com.zanxiang.common.redis.annotation.Lock;
import com.zanxiang.common.redis.exception.LockException;
import com.zanxiang.common.redis.utils.DistributedLockUtil;
import com.zanxiang.common.utils.StringUtils;
import lombok.extern.slf4j.Slf4j;
import org.aspectj.lang.ProceedingJoinPoint;
import org.aspectj.lang.annotation.Around;
import org.aspectj.lang.annotation.Aspect;
import org.aspectj.lang.reflect.MethodSignature;
import org.springframework.context.expression.MethodBasedEvaluationContext;
import org.springframework.core.LocalVariableTableParameterNameDiscoverer;
import org.springframework.expression.ExpressionParser;
import org.springframework.expression.spel.standard.SpelExpressionParser;
import org.springframework.expression.spel.support.StandardEvaluationContext;
import org.springframework.stereotype.Component;

import java.lang.reflect.Method;

/**
 * <p>
 * 锁切面
 * </p>
 *
 * @author zhengwangeng
 * @since 2020/10/19
 */
@Slf4j
@Aspect
@Component
public class LockAspect {
    private static final ExpressionParser EL_PARSER = new SpelExpressionParser();

    @Around("@annotation(com.zanxiang.common.redis.annotation.Lock)")
    public Object around(ProceedingJoinPoint joinPoint) throws Throwable {
        // 获取切入点方法对象
        Method method = ((MethodSignature) joinPoint.getSignature()).getMethod();
        // 从方法上获取自定义注解 Lock, 获取key值， 锁配置等信息
        Lock lockInfo = method.getAnnotation(Lock.class);
        // 获得锁的key
        String lockKey = this.getLockKey(joinPoint, method, lockInfo);
        // 获得锁
        boolean lockResult = DistributedLockUtil.tryLock(lockKey, lockInfo.keepLockTime(), lockInfo.tryLockWaitTime(), lockInfo.timeUnit());
        if (!lockResult) {
            throw new LockException("加锁失败！");
        }
        try {
            return joinPoint.proceed();
        } finally {
            if (lockInfo.autoUnLock()) {
                DistributedLockUtil.unlock(lockKey);
            }
        }
    }


    /**
     * 获取锁的key值
     *
     * @param joinPoint 请求上下文
     * @param method    方法
     * @param lockInfo  锁设置信息
     * @return
     */
    private String getLockKey(ProceedingJoinPoint joinPoint, Method method, Lock lockInfo) {
        if (StringUtils.isBlank(lockInfo.prefix())) {
            throw new LockException("加锁失败！锁前缀不能为空");
        }
        if (StringUtils.isBlank(lockInfo.key())) {
            return lockInfo.prefix();
        }
        // 通过解析字节码文件读取方法的参数名
        LocalVariableTableParameterNameDiscoverer methodParamsDiscover = new LocalVariableTableParameterNameDiscoverer();
        // 实例化spring el上下文
        StandardEvaluationContext context = new MethodBasedEvaluationContext(joinPoint.getTarget(), method, joinPoint.getArgs(), methodParamsDiscover);
        // 获取切入点方法中所有的参数
        String[] params = methodParamsDiscover.getParameterNames(method);
        if (params != null && params.length > 0) {
            // 将切入点方法参数值按顺序放入 el上下文
            for (int i = 0; i < params.length; i++) {
                context.setVariable(params[i], joinPoint.getArgs()[i]);
            }
        }
        String lockKey = EL_PARSER.parseExpression(lockInfo.key()).getValue(context, String.class);
        return lockInfo.prefix() + lockKey;
    }

}
